package net.wanji.opt.service.impl;

import com.github.pagehelper.PageInfo;
import lombok.extern.slf4j.Slf4j;
import net.wanji.common.dto.CrossIdDTO;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.RidInfoMapper;
import net.wanji.feign.pojo.vo.SchemeSendVO;
import net.wanji.opt.dao.mapper.CrossDirInfoMapper;
import net.wanji.opt.dao.mapper.CrossInfoMapper;
import net.wanji.opt.dao.mapper.CrossPhaseMapper;
import net.wanji.opt.dao.mapper.CrossSchemeMapper;
import net.wanji.opt.dao.mapper.CrossSchemeOptLogMapper;
import net.wanji.opt.dao.mapper.strategy.SceneMapper;
import net.wanji.opt.dao.mapper.trend.CrossDataRealtimeMapper;
import net.wanji.opt.dto.CrossIdAndSchemeIdDTO;
import net.wanji.opt.dto.SendManualDTO;
import net.wanji.opt.dto.strategy.AddOrUpdateSceneDTO;
import net.wanji.opt.dto.strategy.QuerySceneDTO;
import net.wanji.opt.po.base.CrossInfoPO;
import net.wanji.opt.po.base.CrossPhasePO;
import net.wanji.opt.po.base.CrossSchemeOptLogPO;
import net.wanji.opt.po.base.CrossSchemePO;
import net.wanji.opt.po.strategy.ScenePO;
import net.wanji.opt.po.trend.CrossDataRealtimePO;
import net.wanji.opt.service.DiagnoService;
import net.wanji.opt.service.signalcontrol.FeignProxyService;
import net.wanji.opt.service.strategy.SceneService;
import net.wanji.opt.vo.CrossIdAndLocationVO;
import net.wanji.opt.vo.SchemeOptSendVO;
import net.wanji.opt.vo.SchemeOptVO;
import org.jetbrains.annotations.NotNull;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2023/3/2 10:09
 */
@Service
@Slf4j
public class DiagnoServiceImpl implements DiagnoService {

    private final CrossInfoMapper crossInfoMapper;
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final SceneService sceneService;
    private final SceneMapper sceneMapper;
    private final CrossDirInfoMapper crossDirInfoMapper;
    private final RidInfoMapper ridInfoMapper;
    private final CrossSchemeOptLogMapper crossSchemeOptLogMapper;
    private final CrossSchemeMapper crossSchemeMapper;
    private final CrossPhaseMapper crossPhaseMapper;
    private final FeignProxyService feignProxyService;

    public DiagnoServiceImpl(CrossInfoMapper crossInfoMapper, CrossDataRealtimeMapper crossDataRealtimeMapper,
                             SceneService sceneService, SceneMapper sceneMapper, CrossDirInfoMapper crossDirInfoMapper,
                             RidInfoMapper ridInfoMapper, CrossSchemeOptLogMapper crossSchemeOptLogMapper,
                             CrossSchemeMapper crossSchemeMapper, CrossPhaseMapper crossPhaseMapper,
                             FeignProxyService feignProxyService) {
        this.crossInfoMapper = crossInfoMapper;
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.sceneService = sceneService;
        this.sceneMapper = sceneMapper;
        this.crossDirInfoMapper = crossDirInfoMapper;
        this.ridInfoMapper = ridInfoMapper;
        this.crossSchemeOptLogMapper = crossSchemeOptLogMapper;
        this.crossSchemeMapper = crossSchemeMapper;
        this.crossPhaseMapper = crossPhaseMapper;
        this.feignProxyService = feignProxyService;
    }

    @Override
    public AddOrUpdateSceneDTO queryCrossScene(CrossIdDTO crossIdDTO) {
        // 查询路口场景代码
        String crossId = crossIdDTO.getCrossId();
        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        Integer status = crossDataRealtimePO.getStatus();
        // 获取场景名称
        ScenePO scenePO = sceneMapper.selectBySceneNum(status);
        String sceneName = scenePO.getSceneName();
        // 构造请求体
        QuerySceneDTO querySceneDTO = new QuerySceneDTO();
        querySceneDTO.setSceneName(sceneName);
        querySceneDTO.setSceneTarget(1);
        querySceneDTO.setPageNum(1);
        querySceneDTO.setPageSize(999);
        PageInfo<AddOrUpdateSceneDTO> pageInfo = sceneService.queryScene(querySceneDTO);
        List<AddOrUpdateSceneDTO> list = pageInfo.getList();
        AddOrUpdateSceneDTO addOrUpdateSceneDTO = list.get(0);

        return addOrUpdateSceneDTO;
    }

    @Override
    public List<CrossIdAndLocationVO> queryNeighborCross(CrossIdDTO crossIdDTO) {
        List<CrossIdAndLocationVO> crossIdAndLocationVOList = new ArrayList<>();
        String crossId = crossIdDTO.getCrossId();
        List<Integer> inDirs = crossDirInfoMapper.selectInDirsByCrossId(crossId);
        for (Integer inDir : inDirs) {
            CrossIdAndLocationVO crossIdAndLocationVO = new CrossIdAndLocationVO();
            RidInfoEntity ridInfo = ridInfoMapper.selectByEndInDir(crossId, inDir);
            if (ridInfo != null) {
                // 关联路口坐标
                String startCrossId = ridInfo.getStartCrossId();
                crossIdAndLocationVO.setCrossId(startCrossId);
                CrossInfoPO crossInfoPO = crossInfoMapper.selectById(startCrossId);
                if (crossInfoPO != null) {
                    String startLocationStr = crossInfoPO.getLocation();
                    double[] startLonLat = CrossUtil.getLonLat(startLocationStr);
                    List<Double> startLocation = new ArrayList<>();
                    startLocation.add(startLonLat[0]);
                    startLocation.add(startLonLat[1]);
                    crossIdAndLocationVO.setLocation(startLocation);
                }
            }
            crossIdAndLocationVOList.add(crossIdAndLocationVO);
        }
        return crossIdAndLocationVOList;
    }

    @Override
    public SchemeOptVO querySchemeOpt(CrossIdDTO crossIdDTO) {
        String crossId = crossIdDTO.getCrossId();
        SchemeOptVO schemeOptVO = new SchemeOptVO();
        schemeOptVO.setCrossId(crossId);
        List<CrossSchemeOptLogPO> crossSchemeOptLogPOList = crossSchemeOptLogMapper.selectByCrossId(crossId);
        // 选择数据批次时间最大的记录
        Optional<CrossSchemeOptLogPO> max = crossSchemeOptLogPOList.stream()
                .max(Comparator.comparing(CrossSchemeOptLogPO::getDataBatchTime));
        List<CrossSchemeOptLogPO> maxList = getMaxList(crossSchemeOptLogPOList, max);
        CrossSchemeOptLogPO crossSchemeOptLogPO = maxList.get(0);
        String optType = crossSchemeOptLogPO.getOptType();
        schemeOptVO.setOptType(Integer.parseInt(optType));
        String schemeNo = crossSchemeOptLogPO.getSchemeNo();
        CrossSchemePO crossSchemePO = crossSchemeMapper.selectByCrossIdAndSchemeNo(crossId, schemeNo);
        schemeOptVO.setCycle(crossSchemePO.getCycle());
        schemeOptVO.setSchemeId(crossSchemePO.getId());
        schemeOptVO.setOffset(crossSchemeOptLogPO.getOffset());
        // 构造调整时间列表
        List<SchemeOptVO.PhaseOptTime> phaseOptTimeList = buildPhaseOptTimeList(maxList);
        schemeOptVO.setPhaseOptTimeList(phaseOptTimeList);
        return schemeOptVO;
    }

    @Override
    public List<SchemeOptVO.PhaseOptTime> queryOptTime(CrossIdAndSchemeIdDTO crossIdAndSchemeIdDTO) {
        String crossId = crossIdAndSchemeIdDTO.getCrossId();
        Integer schemeId = crossIdAndSchemeIdDTO.getSchemeId();
        List<SchemeOptVO.PhaseOptTime> phaseOptTimeList = new ArrayList<>();
        List<CrossSchemeOptLogPO> crossSchemeOptLogPOList = crossSchemeOptLogMapper.selectByCrossId(crossId);
        // 选择数据批次时间最大的记录
        Optional<CrossSchemeOptLogPO> max = crossSchemeOptLogPOList.stream()
                .max(Comparator.comparing(CrossSchemeOptLogPO::getDataBatchTime));
        List<CrossSchemeOptLogPO> maxList = getMaxList(crossSchemeOptLogPOList, max);
        for (CrossSchemeOptLogPO crossSchemeOptLogPO : maxList) {
            SchemeOptVO.PhaseOptTime phaseOptTime = new SchemeOptVO.PhaseOptTime();
            String phaseNo = crossSchemeOptLogPO.getPhaseNo();
            phaseOptTime.setPhaseNo(phaseNo);
            Integer oriGreenTime = crossSchemeOptLogPO.getOriGreenTime();
            phaseOptTime.setOriGreenTime(oriGreenTime);
            Integer optGreenTime = getNewGreenTime(crossId, schemeId, phaseNo);
            phaseOptTime.setOptGreenTime(optGreenTime);
            phaseOptTime.setOptTime(optGreenTime - oriGreenTime);
            phaseOptTimeList.add(phaseOptTime);
        }
        return phaseOptTimeList;
    }

    @Override
    public void sendManual(SendManualDTO sendManualDTO) throws Exception {
        SchemeOptSendVO schemeOptSendVO = new SchemeOptSendVO();
        Integer optTypeInt = sendManualDTO.getOptType();
        String optTypeStr = String.valueOf(optTypeInt);
        schemeOptSendVO.setOptType(optTypeStr);
        Integer crossStatus = sendManualDTO.getCrossStatus();
        String crossStatusStr = String.valueOf(crossStatus);
        schemeOptSendVO.setOptReason(crossStatusStr);

        String currentCrossId = sendManualDTO.getCurrentCrossId();
        String sendCrossId = sendManualDTO.getSendCrossId();
        if (!Objects.equals(currentCrossId, sendCrossId)) {
            schemeOptSendVO.setRelationFlag(String.valueOf(0)); // 是否被关联调整 1否 0是
            schemeOptSendVO.setRelationCrossId(currentCrossId);
        } else {
            schemeOptSendVO.setRelationFlag(String.valueOf(1));
        }

        Map<String, Integer> phaseOffsetTimeMap = new HashMap<>();
        List<SchemeOptVO.PhaseOptTime> phaseOptTimeList = sendManualDTO.getPhaseOptTimeList();
        for (SchemeOptVO.PhaseOptTime phaseOptTime : phaseOptTimeList) {
            phaseOffsetTimeMap.put(phaseOptTime.getPhaseNo(), phaseOptTime.getOptTime());
        }
        schemeOptSendVO.setPhaseOffsetTimeMap(phaseOffsetTimeMap);
        schemeOptSendVO.setCrossCode(sendCrossId);
        Integer schemeId = sendManualDTO.getSchemeId();
        List<SchemeSendVO.Pattern> patternList = buildPatternList(sendCrossId, schemeId, phaseOffsetTimeMap);
        schemeOptSendVO.setPatternList(patternList);

        feignProxyService.schemeOptSend(schemeOptSendVO);
    }

    private List<SchemeSendVO.Pattern> buildPatternList(String sendCrossId, Integer schemeId,
                                                        Map<String, Integer> phaseOffsetTimeMap) {
        List<SchemeSendVO.Pattern> patternList = new ArrayList<>();
        CrossSchemePO crossSchemePO = crossSchemeMapper.selectById(schemeId);
        SchemeSendVO.Pattern pattern = new SchemeSendVO.Pattern();
        pattern.setPatternNo(crossSchemePO.getSchemeNo());
        pattern.setPatternName(crossSchemePO.getName());
        pattern.setCycle(crossSchemePO.getCycle().toString());
        pattern.setCoordPhase(crossSchemePO.getCoordPhase().toString());
        pattern.setOffset(crossSchemePO.getOffset().toString());
        // 构造环列表
        List<SchemeSendVO.Pattern.Ring> rings = buildRings(sendCrossId, schemeId, phaseOffsetTimeMap);
        pattern.setRings(rings);
        patternList.add(pattern);

        return patternList;
    }

    private List<SchemeSendVO.Pattern.Ring> buildRings(String sendCrossId, Integer schemeId,
                                                       Map<String, Integer> phaseOffsetTimeMap) {
        List<SchemeSendVO.Pattern.Ring> rings = new ArrayList<>();
        // 根据路口ID和方案ID查询相位集合，根据ringNo分组，特殊控制ringNo为0
        List<CrossPhasePO> crossPhasePOList = crossPhaseMapper.listCrossPhase(sendCrossId, schemeId.toString());
        Map<Integer, List<CrossPhasePO>> collectMap = crossPhasePOList.stream()
                .collect(Collectors.groupingBy(CrossPhasePO::getRingNo));
        for (Map.Entry<Integer, List<CrossPhasePO>> entry : collectMap.entrySet()) {
            SchemeSendVO.Pattern.Ring ring = new SchemeSendVO.Pattern.Ring();
            ring.setRingNo(entry.getKey().toString());
            // 构造相位列表
            List<CrossPhasePO> phases = entry.getValue();
            List<SchemeSendVO.Pattern.Ring.Phase> phaseList = buildPhaseList(phases, phaseOffsetTimeMap);
            ring.setPhaseList(phaseList);
            rings.add(ring);
        }
        return rings;
    }

    private List<SchemeSendVO.Pattern.Ring.Phase> buildPhaseList(List<CrossPhasePO> phases,
                                                                 Map<String, Integer> phaseOffsetTimeMap) {
        List<SchemeSendVO.Pattern.Ring.Phase> phaseList = new ArrayList<>();
        for (CrossPhasePO phasePO : phases) {
            SchemeSendVO.Pattern.Ring.Phase phase = new SchemeSendVO.Pattern.Ring.Phase();
            String phaseNo = phasePO.getPhaseNo();
            phase.setPhaseNo(phaseNo);
            phase.setPhaseName(phasePO.getName());
            phase.setSort(phasePO.getSort().toString());
            phase.setControlMode(phasePO.getControlMode().toString());
            phase.setMinGreenTime(phasePO.getMinGreenTime().toString());
            phase.setMaxGreenTime(phasePO.getMaxGreenTime().toString());
            phase.setPhaseTime(phasePO.getPhaseTime().toString());

            Integer oldGreenTime = phasePO.getGreenTime();
            Integer optTime = phaseOffsetTimeMap.get(phaseNo);
            if (optTime != null) {
                Integer newGreenTime = oldGreenTime + optTime;
                phase.setGreenTime(newGreenTime.toString());
            } else {
                phase.setGreenTime(oldGreenTime.toString());
            }

            phase.setGreenFlashTime(phasePO.getGreenFlashTime().toString());
            phase.setPedFlashTime(phasePO.getPedFlashTime().toString());
            phase.setYellowTime(phasePO.getYellowTime().toString());
            phase.setRedTime(phasePO.getRedTime().toString());

            phaseList.add(phase);
        }
        return phaseList;
    }

    @Override
    public void restoreSend(CrossIdDTO crossIdDTO) throws Exception {
        String crossId = crossIdDTO.getCrossId();
        feignProxyService.schemeOptRestore(crossId);
    }

    private Integer getNewGreenTime(String crossId, Integer schemeId, String phaseNo) {
        List<CrossPhasePO> crossPhasePOList = crossPhaseMapper.listCrossPhase(crossId, String.valueOf(schemeId));
        for (CrossPhasePO crossPhasePO : crossPhasePOList) {
            String newPhaseNo = crossPhasePO.getPhaseNo();
            if (Objects.equals(phaseNo, newPhaseNo)) {
                return crossPhasePO.getGreenTime();
            }
        }
        return 0;
    }

    @NotNull
    private static List<CrossSchemeOptLogPO> getMaxList(List<CrossSchemeOptLogPO> crossSchemeOptLogPOList,
                                                        Optional<CrossSchemeOptLogPO> max) {
        int maxBatchTime = max.get().getDataBatchTime();
        List<CrossSchemeOptLogPO> maxList = new ArrayList<>();
        for (CrossSchemeOptLogPO crossSchemeOptLogPO : crossSchemeOptLogPOList) {
            int dataBatchTime = crossSchemeOptLogPO.getDataBatchTime();
            if (dataBatchTime == maxBatchTime) {
                maxList.add(crossSchemeOptLogPO);
            }
        }
        return maxList;
    }

    private List<SchemeOptVO.PhaseOptTime> buildPhaseOptTimeList(List<CrossSchemeOptLogPO> crossSchemeOptLogPOList) {
        List<SchemeOptVO.PhaseOptTime> phaseOptTimeList = new ArrayList<>();
        for (CrossSchemeOptLogPO crossSchemeOptLogPO : crossSchemeOptLogPOList) {
            SchemeOptVO.PhaseOptTime phaseOptTime = new SchemeOptVO.PhaseOptTime();
            phaseOptTime.setPhaseNo(crossSchemeOptLogPO.getPhaseNo());
            phaseOptTime.setOriGreenTime(crossSchemeOptLogPO.getOriGreenTime());
            phaseOptTime.setOptGreenTime(crossSchemeOptLogPO.getOptGreenTime());
            phaseOptTime.setOptTime(crossSchemeOptLogPO.getOptTime());
            phaseOptTimeList.add(phaseOptTime);
        }
        return phaseOptTimeList;
    }
}
